from river.multioutput import RegressorChain
from river.metrics.multioutput import MicroAverage
from river.metrics import RMSE
from typing import List, Dict
from movado.estimator import Estimator
from movado.model import Model


class ChainedEstimator(Estimator):
    def __init__(self, model_to_chain: Model, outputs: int):
        super(ChainedEstimator, self).__init__()
        self._chained_model: RegressorChain = RegressorChain(
            model_to_chain.get_model(), order=list(range(outputs))
        )
        self._metric = MicroAverage(RMSE())

    def train(self, X: List[float], y: List[float]) -> None:
        y_true = self.y_to_river(y)
        y_pred = self.y_to_river(self.predict(X))
        if y_pred:
            self._metric.update(y_true, y_pred)
        self._chained_model.learn_one(
            self.X_to_river(X),
            y_true,  # there may be a bug in river typing
        )

    def predict(self, X: List[float]) -> List[float]:
        if not self._chained_model.order:
            self._chained_model.order = list(range(len(X)))
        prediction = self._chained_model.predict_one(
            self.X_to_river(X),
        ).values()  # Here a dict is returned, wrong typing in river 0.7.0
        return prediction

    def get_error(self) -> float:
        return self._metric.get()

    @staticmethod
    def X_to_river(X: List[float]) -> Dict[str, float]:
        return dict(zip(["feature_" + str(i) for i in range(len(X))], X))

    @staticmethod
    def y_to_river(y: List[float]) -> Dict[int, float]:
        return dict(zip(range(len(y)), y))



from abc import ABC
from river.base.regressor import Regressor


class Model(ABC):
    _model: Regressor

    def __init__(self, **kwargs):
        pass

    def get_model(self):
        return self._model



from pathlib import Path
from typing import Dict, Any, List, Callable, Optional, Union

from movado.controller import Controller
from movado.estimator import Estimator
from movado.mab_handler import MabHandler


class ShadowController(Controller):
    def __init__(
        self,
        exact_fitness: Callable[[List[float]], List[float]] = None,
        estimator: Estimator = None,
        self_exact: Optional[object] = None,
        debug: bool = True,
        **kwarg
    ):
        super(ShadowController, self).__init__(
            exact_fitness=exact_fitness,
            estimator=estimator,
            self_exact=self_exact,
            debug=debug,
        )
        if debug:
            self.initialize_debug()

    def compute_objective(
        self, point: List[int], decision_only: bool = False
    ) -> Union[List[float], int]:
        if decision_only:
            out = 1
        else:
            out, exec_time = self._compute_exact(point)
        if self._debug:
            self.write_debug(
                {
                    "Point": point,
                    "Exec_Time": exec_time
                    if not decision_only
                    else "No Exec Time available",
                    "Estimation": 0,
                }
            )
        return out

    def initialize_debug(self):
        Path(self._controller_debug).open("a").write("Point, Exec_Time, Estimation\n")

    def write_debug(self, debug_info: Dict[str, Any]):
        Path(self._controller_debug).open("a").write(
            str(debug_info["Point"])
            + ", "
            + str(debug_info["Exec_Time"])
            + ", "
            + str(debug_info["Estimation"])
            + "\n"
        )

    def get_mab(self) -> MabHandler:
        raise Exception("No Mab Available in Shadow Controller")

    def get_weight_mab(self) -> MabHandler:
        raise Exception("No Weight Mab Available in Shadow Controller")

    def get_parameters(self):
        raise Exception("No Parameters Available for Shadow Controler")



from collections import OrderedDict
from typing import List, Callable, Union
import numpy as np

from movado.controller import Controller
from movado.estimator import Estimator
from functools import wraps
from movado.distance_controller import DistanceController
from movado.hoeffding_adaptive_tree_model import HoeffdingAdaptiveTreeModel
from movado.chained_estimator import ChainedEstimator


# This imports are unused but populate the global symbol table for the globals() call
# noinspection PyUnresolvedReferences
from movado.mab_controller import MabController  # pylint: disable=unused-import

# noinspection PyUnresolvedReferences
from movado.voting_controller import VotingController  # pylint: disable=unused-import

# noinspection PyUnresolvedReferences
from movado.shadow_controller import ShadowController

# noinspection PyUnresolvedReferences
from movado.kernel_regression_model import (
    KernelRegressionModel,
)  # pylint: disable=unused-import


def approximate(
    **kwargs,
) -> Callable[
    [Callable[[List[float]], List[float]]], Callable[[List[float]], List[float]]
]:
    def approximate_decorator(
        func: Callable[[List[float]], List[float]]
    ) -> Callable[[List[float]], List[float]]:
        controller: Controller
        estimator: Estimator
        is_first_call: bool = True
        outputs: int = -1

        @wraps(func)
        def wrapper(*wrapper_args) -> Union[List[float], float]:
            # args must either contain the point to be analyzed
            # or it must contain self (for class methods) and then the point
            nonlocal is_first_call
            nonlocal controller
            nonlocal estimator
            nonlocal outputs

            if len(wrapper_args) < 1 or len(wrapper_args) > 2:
                raise Exception(
                    "The decorated function must have a single input which is a list of numbers, "
                    + "self may also be present for non-static class methods, in such case self is "
                    + "expected to be the first argument followed by the point and nothing"
                )
            if len(wrapper_args) == 1:
                point = wrapper_args[0]
            else:
                point = wrapper_args[1]
            if is_first_call:
                is_first_call = False
                selected_model = globals().get(str(kwargs.get("estimator")) + "Model")
                model = (
                    HoeffdingAdaptiveTreeModel if not selected_model else selected_model
                )
                # TODO add multi-output non-chained estimators
                outputs = kwargs.get("outputs")
                voters: int = kwargs.get("voters")
                if not voters:
                    voters = 750
                if not outputs:
                    raise Exception(
                        "Please specify outputs as a kwarg as the number of targets of the fitness function"
                    )
                estimator = ChainedEstimator(
                    model(
                        **{
                            k[k.find("_") + 1 :]: v
                            for k, v in kwargs.items()
                            if k[: k.find("_")] == "estimator"
                        }
                    ),
                    outputs,
                )
                selected_controller = globals().get(
                    str(kwargs.get("controller")) + "Controller"
                )
                controller = (
                    DistanceController
                    if not selected_controller
                    else selected_controller
                )
                if selected_controller is MabController:
                    params = OrderedDict(
                        {
                            "cover": list(range(3, 20)),
                            "mab_weight_epsilon": [i / 100 for i in range(1, 41)],
                            "mab_weight_bandwidth": list(range(3, 7)),
                        }
                    )
                else:
                    params = OrderedDict(
                        {
                            "nth_nearest": list(range(1, 6)),
                            "mab_epsilon": [i / 100 for i in range(1, 41)],
                            "mab_bandwidth": list(range(3, 7)),
                            "mab_weight_epsilon": [i / 100 for i in range(1, 41)],
                            "mab_weight_bandwidth": list(range(3, 7)),
                        }
                    )
                available_models = int(np.prod([len(v) for v in params.values()]))
                voters = voters if voters <= available_models else available_models
                if voters <= 1:
                    controller = controller(
                        exact_fitness=func,
                        estimator=estimator,
                        self_exact=None if len(wrapper_args) == 1 else wrapper_args[0],
                        **{
                            k[k.find("_") + 1 :]: v
                            for k, v in kwargs.items()
                            if k[: k.find("_")] == "controller"
                        },
                    )
                else:
                    controller = VotingController(
                        controller,
                        estimator,
                        func,
                        voters,
                        params,
                        self_exact=None if len(wrapper_args) == 1 else wrapper_args[0],
                        **{
                            k[k.find("_") + 1 :]: v
                            for k, v in kwargs.items()
                            if k[: k.find("_")] == "controller"
                        },
                    )
            if outputs == 1:
                return list(controller.compute_objective(point))[0]
            return list(controller.compute_objective(point))

        if kwargs.get("disabled"):
            return func
        return wrapper

    return approximate_decorator



from abc import abstractmethod, ABC
from typing import Tuple, Optional, Callable, Any, List, Dict, Union
from sklearn.preprocessing import StandardScaler
import numpy as np
import scipy as sp
import time
from pathlib import Path
from river.preprocessing import StandardScaler as RiverStandardScaler
from river.feature_extraction import RBFSampler
from river.linear_model import LinearRegression
from river.utils import expand_param_grid
from river.model_selection import SuccessiveHalvingRegressor
from river.metrics import RMSE
from collections import deque
from movado.estimator import Estimator
from movado.mab_handler import MabHandler
from movado.chained_estimator import ChainedEstimator

is_call_exact: List[bool] = []  # i-th element True if i-th call is exact


class Controller(ABC):
    @abstractmethod
    def __init__(
        self,
        exact_fitness: Callable[[List[float]], List[float]] = None,
        estimator: Estimator = None,
        self_exact: Optional[object] = None,
        debug: bool = True,
        **kwargs
    ):
        self.__scaler = StandardScaler()
        self.__time_dataset: "deque" = deque()
        self.__error_dataset: "deque" = deque()
        self.__cost_dataset: "deque" = deque()
        self.__to_reshape = False
        self._exact: Callable[[List[float]], List[float]] = exact_fitness
        self._estimator: Estimator = estimator
        self._debug = debug
        self._self_exact = self_exact
        model = RiverStandardScaler()
        model |= RBFSampler(seed=0)
        model |= LinearRegression()
        models = expand_param_grid(
            model,
            {
                "LinearRegression": {
                    "l2": [1e-3, 1e-1, 1, 10],
                },
                "RBFSampler": {"gamma": [1e-3, 1e-1, 1, 10]},
            },
        )
        self.__time_model = SuccessiveHalvingRegressor(
            models=models,
            metric=RMSE(),
            budget=1,
            eta=10000,
            verbose=True,
        )
        self.__error_model = SuccessiveHalvingRegressor(
            models=models,
            metric=RMSE(),
            budget=1,
            eta=10000,
            verbose=True,
        )
        self.__fitted_time: "deque" = deque()
        self.__fitted_error: "deque" = deque()
        self.__fitted_time_mean: float = 0
        self.__fitted_error_mean: float = 0
        self.__cost_mean: float = 0
        self.__time_importance: float = 0
        self.__error_importance: float = 0
        if self._debug:
            Path("movado_debug").mkdir(exist_ok=True)
            self._controller_debug = "movado_debug/controller.csv"
            self.__controller_learn_time_debug = (
                "movado_debug/controller_learn_time.csv"
            )
            Path(self.__controller_learn_time_debug).open("w").close()
            Path(self.__controller_learn_time_debug).open("a").write(
                "Learn Time, Estimation\n"
            )
            Path(self._controller_debug).open("w").close()

    @abstractmethod
    def compute_objective(
        self, point: List[int], decision_only: bool = False
    ) -> List[float]:
        pass

    def learn(
        self,
        is_exact: bool,
        point: List[float],
        exec_time: float = None,
        mab: Optional[Tuple[MabHandler, Union[int, float]]] = None,
        mab_forced_probability: Optional[float] = None,
        mab_forced_action: Optional[Union[int, float]] = None,
        mab_weight: Optional[MabHandler] = None,
        mab_weight_forced_probability: Optional[float] = None,
        mab_weight_forced_action: Optional[Union[int, float]] = None,
        is_point_in_context: bool = True,
    ):

        if mab:
            if mab_weight:
                time_weight = (
                    mab_weight.predict(
                        self._compute_weighting_context(self.get_mab().get_mean_cost())
                    )
                    / 100
                )
                mab[0].learn(
                    cost=self.get_time_error_z_score(
                        time_weight=time_weight, error_weight=1 - time_weight
                    ),
                    context=self._compute_controller_context(point),
                    forced_predict_probability=mab_forced_probability,
                    forced_action=mab_forced_action,
                ),
                mab_weight.learn(
                    cost=self.compute_first_order_sensitivity_index(),
                    context=self._compute_weighting_context(
                        self.get_mab().get_mean_cost()
                    ),
                    forced_predict_probability=mab_weight_forced_probability,
                    forced_action=mab_weight_forced_action,
                )
            else:
                mab[0].learn(
                    cost=self.get_time_error_z_score(),
                    context=self._compute_controller_context(point),
                    forced_predict_probability=mab_forced_probability,
                    forced_action=mab_forced_action,
                )

    def _compute_exact(
        self,
        point: List[int],
        mab: Optional[Tuple[MabHandler, Union[int, float]]] = None,
        mab_forced_probability: Optional[int] = None,
        mab_weight: Optional[MabHandler] = None,
        mab_weight_forced_probability: Optional[int] = None,
        is_point_in_context: bool = True,
    ) -> Tuple[List[float], float]:
        global is_call_exact
        is_call_exact.append(True)
        print("(#" + str(len(is_call_exact)) + ") Waiting for Exact Call...")

        if self._self_exact:
            exact, exec_time = Controller.measure_execution_time(
                self._exact, self._self_exact, point
            )
        else:
            exact, exec_time = Controller.measure_execution_time(self._exact, point)

        self.__time_dataset.append(exec_time)
        self.__error_dataset.append(0)
        _, learn_time = Controller.measure_execution_time(
            self.learn,
            is_exact=True,
            point=point,
            exec_time=exec_time,
            mab=mab,
            mab_forced_probability=mab_forced_probability,
            mab_weight=mab_weight,
            mab_weight_forced_probability=mab_weight_forced_probability,
            is_point_in_context=is_point_in_context,
        )
        Path(self.__controller_learn_time_debug).open("a").write(
            str(learn_time) + "," + str(0) + "\n"
        )
        # self.learn(
        #    is_exact=True,
        #    point=point,
        #    exec_time=exec_time,
        #    mab=mab,
        #    mab_forced_probability=mab_forced_probability,
        #    mab_weight=mab_weight,
        #    mab_weight_forced_probability=mab_weight_forced_probability,
        #    is_point_in_context=is_point_in_context,
        # )
        if np.isscalar(exact):
            exact = [exact]
        self._estimator.train(point, exact)
        return exact, exec_time

    @staticmethod
    def __extract_time_context_features(series: List[float]):
        return (
            (np.mean(series), np.var(series), max(series), min(series), series[-1])
            if len(series) > 0
            else (0, 0, 0, 0, 0)
        )

    @staticmethod
    def _compute_controller_context(point: List[float]):
        return point

    # def _compute_controller_context(self, point: List[int] = None) -> List[float]:
    #     exact_execution_times: List[float] = list(
    #         compress(self.__time_dataset, self.__is_call_exact)
    #     )
    #     estimated_execution_times: List[float] = [
    #         x for x in self.__time_dataset if x not in exact_execution_times
    #     ]
    #     context = [
    #         self._estimator.get_error(),
    #         *self.__extract_time_context_features(exact_execution_times),
    #         *self.__extract_time_context_features(estimated_execution_times),
    #     ]
    #     if not point:
    #         return context
    #     else:
    #         return context + point

    @staticmethod
    def _compute_weighting_context(mab_loss: float) -> List[float]:
        global is_call_exact
        exact_calls = is_call_exact.count(True)
        # return [mab_loss, exact_calls, len(is_call_exact) - exact_calls]
        return [exact_calls, len(is_call_exact) - exact_calls]

    def _compute_estimated(
        self,
        point: List[int],
        mab: Optional[Tuple[MabHandler, Union[int, float]]] = None,
        mab_weight: MabHandler = None,
        is_point_in_context: bool = True,
    ) -> Tuple[List[float], float]:
        global is_call_exact
        is_call_exact.append(False)
        print("(#" + str(len(is_call_exact)) + ") Estimating Exact Call...")

        estimation, exec_time = Controller.measure_execution_time(
            self._estimator.predict,
            point,
        )
        self.__time_dataset.append(exec_time)
        self.__error_dataset.append(self._estimator.get_error())
        _, learn_time = Controller.measure_execution_time(
            self.learn,
            is_exact=False,
            point=point,
            exec_time=exec_time,
            mab=mab,
            mab_forced_probability=None,
            mab_weight=mab_weight,
            mab_weight_forced_probability=None,
            is_point_in_context=is_point_in_context,
        )

        Path(self.__controller_learn_time_debug).open("a").write(
            str(learn_time) + "," + str(1) + "\n"
        )

        return estimation, exec_time

    @abstractmethod
    def initialize_debug(self):
        pass

    @abstractmethod
    def write_debug(self, debug_info: Dict[str, Any]):
        pass

    def __reshape(self, to_reshape):
        if len(self.__time_dataset) > 1:
            return np.reshape(to_reshape, (-1, 1))
        else:
            return np.reshape(to_reshape, (1, -1))

    def get_time_error_z_score(
        self,
        time_weight: float = 1,
        error_weight: float = 1,
    ) -> float:
        reshaped_time = self.__reshape(self.__time_dataset)
        reshaped_error = self.__reshape(self.__error_dataset)
        self.__scaler.fit(reshaped_time)
        time_z_score: float = self.__scaler.transform(reshaped_time)[-1][0]
        self.__scaler.fit(reshaped_error)
        error_z_score: float = self.__scaler.transform(reshaped_error)[-1][0]
        out = (time_weight * time_z_score) + (error_weight * error_z_score)
        self.__cost_dataset.append(out)
        return out

    @staticmethod
    def __is_constant(x):
        return x.count(x[0]) == len(x)

    def get_time_error_correlation(self):
        try:
            samples = len(self.__error_dataset)
            independent_vars = 2
            if (
                self.__is_constant(self.__error_dataset)
                or self.__is_constant(self.__time_dataset)
                or self.__is_constant(self.__cost_dataset)
                or samples - independent_vars == 1
            ):
                return 0
            time_cost_correlation: float = sp.stats.pearsonr(
                self.__time_dataset, self.__cost_dataset
            )[0]
            error_cost_correlation: float = sp.stats.pearsonr(
                self.__error_dataset, self.__cost_dataset
            )[0]
            time_error_correlation: float = sp.stats.pearsonr(
                self.__time_dataset, self.__error_dataset
            )[0]
            R2 = np.sqrt(
                (
                    np.abs(time_cost_correlation**2)
                    + np.abs(error_cost_correlation**2)
                    - 2
                    * time_cost_correlation
                    * error_cost_correlation
                    * time_error_correlation
                )
                / (1 - np.abs(time_error_correlation))
            )
            if R2 == np.nan:
                return 0
            R2_adj = 1 - (
                ((1 - R2**2) * (samples - 1)) / (samples - independent_vars - 1)
            )
            return -R2_adj
        except ValueError:
            return 0

    @staticmethod
    def __update_mean(old_mean: float, old_sample_number: int, new_sample: float):
        """
        Return updated mean in light of new observation without
        full recalculation.
        """
        return (old_sample_number * old_mean + new_sample) / (old_sample_number + 1)

    def compute_first_order_sensitivity_index(self):
        self.__time_model.learn_one(
            ChainedEstimator.X_to_river([self.__time_dataset[-1]]),
            self.__cost_dataset[-1],
        )
        self.__error_model.learn_one(
            ChainedEstimator.X_to_river([self.__error_dataset[-1]]),
            self.__cost_dataset[-1],
        )
        self.__fitted_time.append(
            self.__time_model.predict_one(
                ChainedEstimator.X_to_river([self.__time_dataset[-1]])
            )
        )
        self.__fitted_error.append(
            self.__error_model.predict_one(
                ChainedEstimator.X_to_river([self.__error_dataset[-1]])
            )
        )
        if len(self.__cost_dataset) == 1:
            return 0
        self.__fitted_time_mean = self.__update_mean(
            self.__fitted_time_mean, len(self.__fitted_time) - 1, self.__fitted_time[-1]
        )
        self.__fitted_error_mean = self.__update_mean(
            self.__fitted_error_mean,
            len(self.__fitted_error) - 1,
            self.__fitted_error[-1],
        )
        self.__cost_mean = self.__update_mean(
            self.__cost_mean, len(self.__cost_dataset) - 1, self.__cost_dataset[-1]
        )
        self.__time_importance += (
            self.__fitted_time[-1] - self.__fitted_time_mean
        ) ** 2 / (self.__cost_dataset[-1] - self.__cost_mean) ** 2
        self.__error_importance += (
            self.__fitted_error[-1] - self.__fitted_error_mean
        ) ** 2 / (self.__cost_dataset[-1] - self.__cost_mean) ** 2
        importance_sum = self.__time_importance + self.__error_importance
        time_importance = self.__time_importance / importance_sum
        error_importance = self.__error_importance / importance_sum
        return (0.5 - time_importance) ** 2 + (0.5 - error_importance) ** 2

    @staticmethod
    def measure_execution_time(
        func: Callable[..., Any], *args, **kwargs
    ) -> Tuple[Any, float]:
        start = time.time()
        out = func(*args, **kwargs)
        execution_time = time.time() - start
        return out, execution_time

    @abstractmethod
    def get_mab(self) -> MabHandler:
        pass

    @abstractmethod
    def get_weight_mab(self) -> MabHandler:
        pass

    @abstractmethod
    def get_parameters(self):
        pass



import river as rv
from river.preprocessing import StandardScaler
from river.feature_extraction import RBFSampler
from river.linear_model import LinearRegression
from river.utils import expand_param_grid

from movado.model import Model


class KernelRegressionModel(Model):
    def __init__(
        self,
        eta=2,
        budget=2000,
    ):
        super(KernelRegressionModel, self).__init__()
        self._model = StandardScaler()
        self._model |= RBFSampler()

        self._model |= LinearRegression()

        models = expand_param_grid(
            self._model,
            {
                "LinearRegression": {
                    "optimizer": [
                        (
                            rv.optim.Adam,
                            {
                                "beta_1": [0.1, 0.01, 0.001],
                                "lr": [0.1, 0.01, 0.001],
                            },
                        )
                    ],
                    "l2": [
                        0.001,
                        0.002,
                        0.004,
                        0.008,
                        0.016,
                        0.032,
                        0.064,
                        0.128,
                        0.256,
                        0.512,
                        0.1024,
                        0.2048,
                        0.4096,
                    ],
                    "intercept_lr": [
                        rv.optim.schedulers.Optimal(loss=rv.optim.losses.Squared())
                    ],
                    "initializer": [rv.optim.initializers.Normal()],
                },
                "RBFSampler": {"gamma": [1e-3, 1e-1, 1, 10]},
            },
        )

        self._model = rv.expert.SuccessiveHalvingRegressor(
            models=models,
            metric=rv.metrics.MAE(),
            budget=budget,
            eta=eta,
            verbose=True,
        )



from river.preprocessing import StandardScaler
from river.feature_extraction import RBFSampler
from river.tree.hoeffding_adaptive_tree_regressor import HoeffdingAdaptiveTreeRegressor
from river.model_selection import SuccessiveHalvingRegressor
from river.utils import expand_param_grid
from river.metrics import RMSE

from movado.model import Model


class HoeffdingAdaptiveTreeModel(Model):
    def __init__(
        self,
        eta=2,
        budget=2000,
    ):
        super(HoeffdingAdaptiveTreeModel, self).__init__()
        self._model = StandardScaler()
        self._model |= RBFSampler(seed=0)

        self._model |= HoeffdingAdaptiveTreeRegressor(
            leaf_prediction="adaptive", seed=0
        )
        models = expand_param_grid(
            self._model,
            {
                "HoeffdingAdaptiveTreeRegressor": {
                    "model_selector_decay": [
                        0.35,
                        0.55,
                        0.75,
                        0.95,
                    ],
                },
                "RBFSampler": {"gamma": [1e-3, 1e-1, 1, 10]},
            },
        )
        self._model = SuccessiveHalvingRegressor(
            models=models,
            metric=RMSE(),
            budget=budget,
            eta=eta,
            verbose=True,
        )
